@@ -28,3 +28,6 @@ pub mod exec;
 mod lexer;
 mod parser;
 mod reader;
+pub mod syms;
+#[cfg(test)]
+mod testutils;
