@@ -46,7 +46,6 @@ pub type Result<T> = std::result::Result<T, Error>;
 #[derive(Debug, Eq, PartialEq)]
 enum ExprOp {
     LeftParen,
-    Call(VarRef),
 
     Add,
     Subtract,
@@ -95,7 +94,6 @@ impl ExprOp {
     fn priority(&self) -> i8 {
         match self {
             ExprOp::LeftParen => 5,
-            ExprOp::Call(_) => 5,
 
             ExprOp::Negate => 4,
             ExprOp::Not => 4,
@@ -161,7 +159,6 @@ impl ExprOp {
             ExprOp::Not => apply1(exprs, Expr::Not),
 
             ExprOp::LeftParen => Ok(()),
-            ExprOp::Call(_) => Ok(()),
         }
     }
 }
@@ -210,7 +207,7 @@ impl<'a> Parser<'a> {
 
     /// Parses an assignment for the variable reference `varref` already read.
     fn parse_assignment(&mut self, vref: VarRef) -> Result<Statement> {
-        let expr = match self.parse_expr()? {
+        let expr = match self.parse_expr(None)? {
             Some(expr) => expr,
             None => return Err(Error::Bad("Missing expression in assignment".to_owned())),
         };
@@ -223,8 +220,24 @@ impl<'a> Parser<'a> {
         Ok(Statement::Assignment(vref, expr))
     }
 
+    /// Parses an assignment to the array `varref` with `subscripts`, both of which have already
+    /// been read.
+    fn parse_array_assignment(&mut self, vref: VarRef, subscripts: Vec<Expr>) -> Result<Statement> {
+        let expr = match self.parse_expr(None)? {
+            Some(expr) => expr,
+            None => return Err(Error::Bad("Missing expression in array assignment".to_owned())),
+        };
+
+        let next = self.lexer.peek()?;
+        match next {
+            Token::Eof | Token::Eol => (),
+            _ => return Err(Error::Bad("Unexpected token in array assignment".to_owned())),
+        }
+        Ok(Statement::ArrayAssignment(vref, subscripts, expr))
+    }
+
     /// Parses a builtin call (things of the form `INPUT a`).
-    fn parse_builtin_call(&mut self, vref: VarRef) -> Result<Statement> {
+    fn parse_builtin_call(&mut self, vref: VarRef, mut first: Option<Expr>) -> Result<Statement> {
         let mut name = match vref.into_unannotated_string() {
             Ok(name) => name,
             Err(e) => return Err(Error::Bad(format!("{}", e))),
@@ -233,7 +246,7 @@ impl<'a> Parser<'a> {
 
         let mut args = vec![];
         loop {
-            let expr = self.parse_expr()?;
+            let expr = self.parse_expr(first.take())?;
 
             let peeked = self.lexer.peek()?;
             match peeked {
@@ -261,19 +274,144 @@ impl<'a> Parser<'a> {
         Ok(Statement::BuiltinCall(name, args))
     }
 
+    /// Starts processing either an array reference or a builtin call and disambiguates between the
+    /// two.
+    fn parse_array_or_builtin_call(&mut self, vref: VarRef) -> Result<Statement> {
+        match self.lexer.peek()? {
+            Token::LeftParen => {
+                self.lexer.consume_peeked();
+                let mut exprs = self.parse_comma_separated_exprs()?;
+                match self.lexer.peek()? {
+                    Token::Equal => {
+                        self.lexer.consume_peeked();
+                        self.parse_array_assignment(vref, exprs)
+                    }
+                    _ => {
+                        if exprs.len() != 1 {
+                            return Err(Error::Bad("Expected expression".to_owned()));
+                        }
+                        self.parse_builtin_call(vref, Some(exprs.remove(0)))
+                    }
+                }
+            }
+            _ => self.parse_builtin_call(vref, None),
+        }
+    }
+
+    /// Parses the `AS typename` clause of a `DIM` statement.  The caller has already consumed the
+    /// `AS` token.
+    fn parse_dim_as(&mut self) -> Result<VarType> {
+        let peeked = self.lexer.peek()?;
+        let vartype = match peeked {
+            Token::Eof | Token::Eol => VarType::Integer,
+            Token::As => {
+                self.lexer.consume_peeked();
+                match self.lexer.read()? {
+                    Token::BooleanName => VarType::Boolean,
+                    Token::DoubleName => VarType::Double,
+                    Token::IntegerName => VarType::Integer,
+                    Token::TextName => VarType::Text,
+                    t => {
+                        return Err(Error::Bad(format!(
+                            "Invalid type name {:?} in DIM AS statement",
+                            t
+                        )))
+                    }
+                }
+            }
+            _ => return Err(Error::Bad("Expected AS or end of statement".to_owned())),
+        };
+
+        let next = self.lexer.peek()?;
+        match next {
+            Token::Eof | Token::Eol => (),
+            _ => return Err(Error::Bad("Unexpected token in DIM statement".to_owned())),
+        }
+
+        Ok(vartype)
+    }
+
+    /// Parses a `DIM` statement.
+    fn parse_dim(&mut self) -> Result<Statement> {
+        let vref = match self.lexer.read()? {
+            Token::Symbol(vref) => vref,
+            _ => return Err(Error::Bad("Expected variable name after DIM".to_owned())),
+        };
+        let name = vref.into_unannotated_string()?;
+
+        let peeked = self.lexer.peek()?;
+        match peeked {
+            Token::LeftParen => {
+                self.lexer.consume_peeked();
+                let subscripts = self.parse_comma_separated_exprs()?;
+                if subscripts.is_empty() {
+                    return Err(Error::Bad("Arrays require at least one dimension".to_owned()));
+                }
+                let vartype = self.parse_dim_as()?;
+                Ok(Statement::DimArray(name, subscripts, vartype))
+            }
+            _ => {
+                let vartype = self.parse_dim_as()?;
+                Ok(Statement::Dim(name, vartype))
+            }
+        }
+    }
+
+    /// Parses a variable list of comma-separated expressions.  The caller must have consumed the
+    /// open parenthesis and we stop processing when we encounter the terminating parenthesis (and
+    /// consume it).  We expect at least one expression.
+    fn parse_comma_separated_exprs(&mut self) -> Result<Vec<Expr>> {
+        let mut exprs = vec![];
+        if let Some(expr) = self.parse_expr(None)? {
+            // The first expression is optional to support calls to functions without arguments.
+            exprs.push(expr);
+        }
+
+        loop {
+            let peeked = self.lexer.peek()?;
+            match peeked {
+                Token::RightParen => {
+                    self.lexer.consume_peeked();
+                    break;
+                }
+                Token::Comma => {
+                    self.lexer.consume_peeked();
+                    if exprs.is_empty() {
+                        // The first expression (parsed outside the loop) cannot be empty if we
+                        // encounter more than one expression.
+                        return Err(Error::Bad("Missing expression".to_owned()));
+                    }
+                    match self.parse_expr(None)? {
+                        Some(expr) => exprs.push(expr),
+                        None => return Err(Error::Bad("Missing expression".to_owned())),
+                    }
+                }
+                _ => return Err(Error::Bad("Unexpected token".to_owned())),
+            }
+        }
+
+        Ok(exprs)
+    }
+
     /// Parses an expression.
     ///
-    /// It is important to single out the special `Empty` expression as a possible return value,
-    /// which denotes that no expression was found.  This is necessary to treat the case of empty
+    /// Returns `None` if no expression was found.  This is necessary to treat the case of empty
     /// arguments to statements, as is the case in `PRINT a , , b`.
     ///
+    /// If the caller has already processed a parenthesized term of an expression like
+    /// `(first) + second`, then that term must be provided in `first`.
+    ///
     /// This is an implementation of the Shunting Yard Algorithm by Edgar Dijkstra.
-    fn parse_expr(&mut self) -> Result<Option<Expr>> {
+    fn parse_expr(&mut self, first: Option<Expr>) -> Result<Option<Expr>> {
         let mut exprs: Vec<Expr> = vec![];
         let mut ops: Vec<ExprOp> = vec![];
-        let mut arity: Vec<usize> = vec![];
 
         let mut need_operand = true; // Also tracks whether an upcoming minus is unary.
+        if let Some(e) = first {
+            exprs.push(e);
+            need_operand = false;
+        }
+
         loop {
             let mut handle_operand = |e| {
                 if !need_operand {
@@ -289,15 +427,18 @@ impl<'a> Parser<'a> {
             match self.lexer.peek()? {
                 Token::Eof
                 | Token::Eol
+                | Token::Comma
                 | Token::Semicolon
                 | Token::Then
                 | Token::To
                 | Token::Step => break,
-                Token::Comma => {
-                    // The comma is a special separator: if it's outside of a function call, the
-                    // expression is complete.  But if it is within a function call, we need to
-                    // continue our processing to extract function arguments.
-                    if arity.is_empty() {
+                Token::RightParen => {
+                    if !ops.contains(&ExprOp::LeftParen) {
+                        // We encountered an unbalanced parenthesis but we don't know if this is
+                        // because we were called from within an argument list (in which case the
+                        // caller consumed the opening parenthesis and is expecting to consume the
+                        // closing parenthesis) or because we really found an invalid expression.
+                        // Only the caller can know, so avoid consuming the token and exit.
                         break;
                     }
                 }
@@ -317,9 +458,18 @@ impl<'a> Parser<'a> {
                     // parenthesis into the beginning of a function call.
                     match exprs.pop() {
                         Some(Expr::Symbol(vref)) => {
-                            debug_assert!(!need_operand);
-                            ops.push(ExprOp::Call(vref));
-                            arity.push(exprs.len());
+                            if !need_operand {
+                                exprs.push(Expr::Call(vref, self.parse_comma_separated_exprs()?));
+                                need_operand = false;
+                            } else {
+                                // We popped out the last expression to see if it this left
+                                // parenthesis started a function call... but it did not (it is a
+                                // symbol following a parenthesis) so put both the expression and
+                                // the token back.
+                                ops.push(ExprOp::LeftParen);
+                                exprs.push(Expr::Symbol(vref));
+                                need_operand = true;
+                            }
                         }
                         e => {
                             if let Some(e) = e {
@@ -334,47 +484,21 @@ impl<'a> Parser<'a> {
                                 ));
                             }
                             ops.push(ExprOp::LeftParen);
+                            need_operand = true;
                         }
                     };
-                    need_operand = true;
                 }
                 Token::RightParen => {
+                    let mut found = false;
                     while let Some(op) = ops.pop() {
                         op.apply(&mut exprs)?;
-                        match op {
-                            ExprOp::LeftParen | ExprOp::Call(_) => {
-                                ops.push(op);
-                                break;
-                            }
-                            _ => (),
-                        };
-                    }
-
-                    match ops.pop() {
-                        Some(ExprOp::LeftParen) => {}
-                        Some(ExprOp::Call(name)) => {
-                            let n = exprs.len() - arity.pop().unwrap();
-                            let params = exprs.split_off(exprs.len() - n);
-                            exprs.push(Expr::Call(name, params));
-                        }
-                        Some(_) => (),
-                        None => return Err(Error::Bad("Unbalanced parenthesis".to_owned())),
-                    }
-                    need_operand = false;
-                }
-                Token::Comma => {
-                    debug_assert!(!ops.is_empty());
-                    if need_operand {
-                        return Err(Error::Bad("Missing expression after comma".to_owned()));
-                    }
-                    while let Some(op) = ops.pop() {
-                        op.apply(&mut exprs)?;
-                        if let ExprOp::Call(_) = op {
-                            ops.push(op);
+                        if op == ExprOp::LeftParen {
+                            found = true;
                             break;
                         }
                     }
-                    need_operand = true;
+                    assert!(found, "Unbalanced parenthesis should have been handled above");
+                    need_operand = false;
                 }
 
                 Token::Not => {
@@ -388,11 +512,9 @@ impl<'a> Parser<'a> {
                     } else {
                         op = ExprOp::Subtract;
                         while let Some(op2) = ops.last() {
-                            match *op2 {
-                                ExprOp::LeftParen | ExprOp::Call(_) => break,
-                                _ if op2.priority() < op.priority() => break,
-                                _ => (),
-                            };
+                            if *op2 == ExprOp::LeftParen || op2.priority() < op.priority() {
+                                break;
+                            }
                             let op2 = ops.pop().unwrap();
                             op2.apply(&mut exprs)?;
                         }
@@ -416,11 +538,9 @@ impl<'a> Parser<'a> {
                 | Token::Xor => {
                     let op = ExprOp::from(token);
                     while let Some(op2) = ops.last() {
-                        match *op2 {
-                            ExprOp::LeftParen | ExprOp::Call(_) => break,
-                            _ if op2.priority() < op.priority() => break,
-                            _ => (),
-                        };
+                        if *op2 == ExprOp::LeftParen || op2.priority() < op.priority() {
+                            break;
+                        }
                         let op2 = ops.pop().unwrap();
                         op2.apply(&mut exprs)?;
                     }
@@ -432,6 +552,7 @@ impl<'a> Parser<'a> {
 
                 Token::Eof
                 | Token::Eol
+                | Token::Comma
                 | Token::Semicolon
                 | Token::Then
                 | Token::To
@@ -439,13 +560,20 @@ impl<'a> Parser<'a> {
                     panic!("Field separators handled above")
                 }
 
-                Token::If
+                Token::As
+                | Token::BooleanName
+                | Token::Dim
+                | Token::DoubleName
                 | Token::Else
                 | Token::Elseif
                 | Token::End
-                | Token::While
                 | Token::For
-                | Token::Next => {
+                | Token::If
+                | Token::IntegerName
+                | Token::Next
+                | Token::TextName
+                | Token::Wend
+                | Token::While => {
                     return Err(Error::Bad("Unexpected keyword in expression".to_owned()));
                 }
             };
@@ -467,7 +595,7 @@ impl<'a> Parser<'a> {
 
     /// Parses an `IF` statement.
     fn parse_if(&mut self) -> Result<Statement> {
-        let expr = match self.parse_expr()? {
+        let expr = match self.parse_expr(None)? {
             Some(expr) => expr,
             None => return Err(Error::Bad("No expression in IF statement".to_owned())),
         };
@@ -481,7 +609,7 @@ impl<'a> Parser<'a> {
             match peeked {
                 Token::Elseif => {
                     self.lexer.consume_peeked();
-                    let expr = match self.parse_expr()? {
+                    let expr = match self.parse_expr(None)? {
                         Some(expr) => expr,
                         None => {
                             return Err(Error::Bad("No expression in ELSEIF statement".to_owned()))
@@ -578,13 +706,13 @@ impl<'a> Parser<'a> {
         };
 
         self.expect_and_consume(Token::Equal, "No equal sign in FOR statement")?;
-        let start = match self.parse_expr()? {
+        let start = match self.parse_expr(None)? {
             Some(expr) => expr,
             None => return Err(Error::Bad("No start expression in FOR statement".to_owned())),
         };
 
         self.expect_and_consume(Token::To, "No TO in FOR statement")?;
-        let end = match self.parse_expr()? {
+        let end = match self.parse_expr(None)? {
             Some(expr) => expr,
             None => return Err(Error::Bad("No end expression in FOR statement".to_owned())),
         };
@@ -632,15 +760,14 @@ impl<'a> Parser<'a> {
 
     /// Parses a `WHILE` statement.
     fn parse_while(&mut self) -> Result<Statement> {
-        let expr = match self.parse_expr()? {
+        let expr = match self.parse_expr(None)? {
             Some(expr) => expr,
             None => return Err(Error::Bad("No expression in WHILE statement".to_owned())),
         };
         self.expect_and_consume(Token::Eol, "Expecting newline after WHILE")?;
 
-        let stmts = self.parse_until(&[Token::End])?;
-        self.expect_and_consume(Token::End, "WHILE without END WHILE")?;
-        self.expect_and_consume(Token::While, "WHILE without END WHILE")?;
+        let stmts = self.parse_until(&[Token::Wend])?;
+        self.expect_and_consume(Token::Wend, "WHILE without WEND")?;
 
         Ok(Statement::While(expr, stmts))
     }
@@ -652,7 +779,7 @@ impl<'a> Parser<'a> {
                 Token::Eof => break,
                 Token::End => {
                     self.lexer.consume_peeked();
-                    self.expect_and_consume(Token::While, "WHILE without END WHILE")?;
+                    self.expect_and_consume(Token::While, "WHILE without WEND")?;
                     break;
                 }
                 _ => {
@@ -680,6 +807,7 @@ impl<'a> Parser<'a> {
         let res = match self.lexer.read()? {
             Token::Eof => return Ok(None),
             Token::Eol => Ok(None),
+            Token::Dim => Ok(Some(self.parse_dim()?)),
             Token::If => {
                 let result = self.parse_if();
                 if result.is_err() {
@@ -700,7 +828,7 @@ impl<'a> Parser<'a> {
                     self.lexer.consume_peeked();
                     Ok(Some(self.parse_assignment(vref)?))
                 } else {
-                    Ok(Some(self.parse_builtin_call(vref)?))
+                    Ok(Some(self.parse_array_or_builtin_call(vref)?))
                 }
             }
             Token::While => {
@@ -813,6 +941,49 @@ mod tests {
         );
     }
 
+    #[test]
+    fn test_array_assignments() {
+        do_ok_test(
+            "a(1)=100\nfoo(2, 3)=\"text\"\nabc$ (5 + z, 6) = TRUE OR FALSE",
+            &[
+                Statement::ArrayAssignment(
+                    VarRef::new("a", VarType::Auto),
+                    vec![Expr::Integer(1)],
+                    Expr::Integer(100),
+                ),
+                Statement::ArrayAssignment(
+                    VarRef::new("foo", VarType::Auto),
+                    vec![Expr::Integer(2), Expr::Integer(3)],
+                    Expr::Text("text".to_owned()),
+                ),
+                Statement::ArrayAssignment(
+                    VarRef::new("abc", VarType::Text),
+                    vec![
+                        Expr::Add(
+                            Box::from(Expr::Integer(5)),
+                            Box::from(Expr::Symbol(VarRef::new("z".to_owned(), VarType::Auto))),
+                        ),
+                        Expr::Integer(6),
+                    ],
+                    Expr::Or(Box::from(Expr::Boolean(true)), Box::from(Expr::Boolean(false))),
+                ),
+            ],
+        );
+    }
+
+    #[test]
+    fn test_array_assignment_errors() {
+        do_error_test("a(", "Unexpected token");
+        do_error_test("a()", "Expected expression");
+        do_error_test("a() =", "Missing expression in array assignment");
+        do_error_test("a() IF", "Expected expression");
+        do_error_test("a() = 3 4", "Unexpected value in expression");
+        do_error_test("a() = 3 THEN", "Unexpected token in array assignment");
+        do_error_test("a(,) = 3", "Missing expression");
+        do_error_test("a(2;3) = 3", "Unexpected token");
+        do_error_test("(2) = 3", "Unexpected token LeftParen in statement");
+    }
+
     #[test]
     fn test_assignments() {
         do_ok_test(
@@ -866,6 +1037,51 @@ mod tests {
         );
     }
 
+    #[test]
+    fn test_builtin_calls_and_array_references_disambiguation() {
+        use Expr::*;
+
+        do_ok_test(
+            "PRINT(1)",
+            &[Statement::BuiltinCall("PRINT".to_owned(), vec![(Some(Integer(1)), ArgSep::End)])],
+        );
+
+        do_ok_test(
+            "PRINT(1), 2",
+            &[Statement::BuiltinCall(
+                "PRINT".to_owned(),
+                vec![(Some(Integer(1)), ArgSep::Long), (Some(Integer(2)), ArgSep::End)],
+            )],
+        );
+
+        do_ok_test(
+            "PRINT(1); 2",
+            &[Statement::BuiltinCall(
+                "PRINT".to_owned(),
+                vec![(Some(Integer(1)), ArgSep::Short), (Some(Integer(2)), ArgSep::End)],
+            )],
+        );
+
+        do_ok_test(
+            "PRINT(1);",
+            &[Statement::BuiltinCall(
+                "PRINT".to_owned(),
+                vec![(Some(Integer(1)), ArgSep::Short), (None, ArgSep::End)],
+            )],
+        );
+
+        do_ok_test(
+            "PRINT(1) + 2; 3",
+            &[Statement::BuiltinCall(
+                "PRINT".to_owned(),
+                vec![
+                    (Some(Add(Box::from(Integer(1)), Box::from(Integer(2)))), ArgSep::Short),
+                    (Some(Integer(3)), ArgSep::End),
+                ],
+            )],
+        );
+    }
+
     #[test]
     fn test_builtin_calls_error() {
         do_error_test("FOO 3 5\n", "Unexpected value in expression");
@@ -873,6 +1089,88 @@ mod tests {
         do_error_test("PRINT IF 1\n", "Unexpected keyword in expression");
         do_error_test("PRINT 3, IF 1\n", "Unexpected keyword in expression");
         do_error_test("PRINT 3 THEN\n", "Expected comma, semicolon, or end of statement");
+        do_error_test("PRINT ()\n", "Expected expression");
+        do_error_test("PRINT (2, 3)\n", "Expected expression");
+        do_error_test("PRINT (2, 3); 4\n", "Expected expression");
+    }
+
+    #[test]
+    fn test_dim_default_type() {
+        do_ok_test("DIM i", &[Statement::Dim("i".to_owned(), VarType::Integer)]);
+    }
+
+    #[test]
+    fn test_dim_as_simple_types() {
+        do_ok_test("DIM i AS BOOLEAN", &[Statement::Dim("i".to_owned(), VarType::Boolean)]);
+        do_ok_test("DIM i AS DOUBLE", &[Statement::Dim("i".to_owned(), VarType::Double)]);
+        do_ok_test("DIM i AS INTEGER", &[Statement::Dim("i".to_owned(), VarType::Integer)]);
+        do_ok_test("DIM i AS STRING", &[Statement::Dim("i".to_owned(), VarType::Text)]);
+    }
+
+    #[test]
+    fn test_dim_consecutive() {
+        do_ok_test(
+            "DIM i\nDIM j AS BOOLEAN\nDIM k",
+            &[
+                Statement::Dim("i".to_owned(), VarType::Integer),
+                Statement::Dim("j".to_owned(), VarType::Boolean),
+                Statement::Dim("k".to_owned(), VarType::Integer),
+            ],
+        );
+    }
+
+    #[test]
+    fn test_dim_array() {
+        use Expr::*;
+
+        do_ok_test(
+            "DIM i(10)",
+            &[Statement::DimArray("i".to_owned(), vec![Integer(10)], VarType::Integer)],
+        );
+
+        do_ok_test(
+            "DIM foo(-5, 0) AS STRING",
+            &[Statement::DimArray(
+                "foo".to_owned(),
+                vec![Negate(Box::from(Integer(5))), Integer(0)],
+                VarType::Text,
+            )],
+        );
+
+        do_ok_test(
+            "DIM foo(bar$() + 3, 8, -1)",
+            &[Statement::DimArray(
+                "foo".to_owned(),
+                vec![
+                    Add(
+                        Box::from(Call(VarRef::new("bar", VarType::Text), vec![])),
+                        Box::from(Integer(3)),
+                    ),
+                    Integer(8),
+                    Negate(Box::from(Integer(1))),
+                ],
+                VarType::Integer,
+            )],
+        );
+    }
+
+    #[test]
+    fn test_dim_errors() {
+        do_error_test("DIM", "Expected variable name after DIM");
+        do_error_test("DIM 3", "Expected variable name after DIM");
+        do_error_test("DIM AS", "Expected variable name after DIM");
+        do_error_test("DIM foo 3", "Expected AS or end of statement");
+        do_error_test("DIM a AS", "Invalid type name Eof in DIM AS statement");
+        do_error_test("DIM a$ AS", "Type annotation not allowed in a$");
+        do_error_test("DIM a AS 3", "Invalid type name Integer(3) in DIM AS statement");
+        do_error_test("DIM a AS INTEGER 3", "Unexpected token in DIM statement");
+
+        do_error_test("DIM a()", "Arrays require at least one dimension");
+        do_error_test("DIM a(,)", "Missing expression");
+        do_error_test("DIM a(, 3)", "Missing expression");
+        do_error_test("DIM a(3, )", "Missing expression");
+        do_error_test("DIM a(3, , 4)", "Missing expression");
+        do_error_test("DIM a(1) AS INTEGER 3", "Unexpected token in DIM statement");
     }
 
     /// Wrapper around `do_ok_test` to parse an expression.  Given that expressions alone are not
@@ -924,6 +1222,13 @@ mod tests {
                 Box::from(Add(Box::from(Integer(2)), Box::from(Integer(5)))),
             ),
         );
+        do_expr_ok_test(
+            "(7) - (1) + (-2)",
+            Add(
+                Box::from(Subtract(Box::from(Integer(7)), Box::from(Integer(1)))),
+                Box::from(Negate(Box::from(Integer(2)))),
+            ),
+        );
     }
 
     #[test]
@@ -1043,6 +1348,13 @@ mod tests {
     #[test]
     fn test_expr_functions_nested() {
         use Expr::*;
+        do_expr_ok_test(
+            "consecutive(parenthesis())",
+            Call(
+                VarRef::new("consecutive", VarType::Auto),
+                vec![Call(VarRef::new("parenthesis", VarType::Auto), vec![])],
+            ),
+        );
         do_expr_ok_test(
             "outer?(1, inner1(2, 3), 4, inner2(), 5)",
             Call(
@@ -1062,7 +1374,7 @@ mod tests {
     fn test_expr_functions_and_ops() {
         use Expr::*;
         do_expr_ok_test(
-            "b AND ask?(34 + 15, ask(1, FALSE))",
+            "b AND ask?(34 + 15, ask(1, FALSE), -5)",
             And(
                 Box::from(Symbol(VarRef::new("b".to_owned(), VarType::Auto))),
                 Box::from(Call(
@@ -1070,12 +1382,30 @@ mod tests {
                     vec![
                         Add(Box::from(Integer(34)), Box::from(Integer(15))),
                         Call(VarRef::new("ask", VarType::Auto), vec![Integer(1), Boolean(false)]),
+                        Negate(Box::from(Integer(5))),
                     ],
                 )),
             ),
         );
     }
 
+    #[test]
+    fn test_expr_functions_not_confused_with_symbols() {
+        use Expr::*;
+        let iref = VarRef::new("i", VarType::Auto);
+        let jref = VarRef::new("j", VarType::Auto);
+        do_expr_ok_test(
+            "i = 0 OR i = (j - 1)",
+            Or(
+                Box::from(Equal(Box::from(Symbol(iref.clone())), Box::from(Integer(0)))),
+                Box::from(Equal(
+                    Box::from(Symbol(iref)),
+                    Box::from(Subtract(Box::from(Symbol(jref)), Box::from(Integer(1)))),
+                )),
+            ),
+        );
+    }
+
     #[test]
     fn test_expr_errors() {
         do_expr_error_test("+3", "Not enough values to apply operator");
@@ -1084,19 +1414,24 @@ mod tests {
         do_expr_error_test("((3)2)", "Unexpected value in expression");
         do_expr_error_test("2 3", "Unexpected value in expression");
 
-        do_expr_error_test("(", "Unbalanced parenthesis");
-        do_expr_error_test(")", "Unbalanced parenthesis");
-        do_expr_error_test("(()", "Unbalanced parenthesis");
-        do_expr_error_test("())", "Unbalanced parenthesis");
+        do_expr_error_test("(", "Missing expression");
+        do_expr_error_test(")", "Expected comma, semicolon, or end of statement");
+        do_expr_error_test("(()", "Missing expression");
+        do_expr_error_test("())", "Expected expression");
         do_expr_error_test("3 + (2 + 1) + (4 - 5", "Unbalanced parenthesis");
-        do_expr_error_test("3 + 2 + 1) + (4 - 5)", "Unbalanced parenthesis");
+        do_expr_error_test(
+            "3 + 2 + 1) + (4 - 5)",
+            "Expected comma, semicolon, or end of statement",
+        );
 
-        do_expr_error_test("foo(,)", "Missing expression after comma");
-        do_expr_error_test("foo(3, , 4)", "Missing expression after comma");
+        do_expr_error_test("foo(,)", "Missing expression");
+        do_expr_error_test("foo(, 3)", "Missing expression");
+        do_expr_error_test("foo(3, )", "Missing expression");
+        do_expr_error_test("foo(3, , 4)", "Missing expression");
         // TODO(jmmv): These are not the best error messages...
-        do_expr_error_test("(,)", "Unbalanced parenthesis");
-        do_expr_error_test("(3, 4)", "Unbalanced parenthesis");
-        do_expr_error_test("((), ())", "Unbalanced parenthesis");
+        do_expr_error_test("(,)", "Missing expression");
+        do_expr_error_test("(3, 4)", "Expected expression");
+        do_expr_error_test("((), ())", "Missing expression");
 
         // TODO(jmmv): This succeeds because `PRINT` is interned as a `Token::Symbol` so the
         // expression parser sees it as a variable reference... but this should probably fail.
@@ -1111,7 +1446,10 @@ mod tests {
 
     #[test]
     fn test_expr_errors_due_to_keywords() {
-        for kw in &["IF", "ELSEIF", "ELSE", "END", "WHILE", "FOR", "NEXT"] {
+        for kw in &[
+            "AS", "BOOLEAN", "DIM", "DOUBLE", "ELSE", "ELSEIF", "END", "FOR", "IF", "INTEGER",
+            "NEXT", "STRING", "WHILE",
+        ] {
             do_expr_error_test(&format!("2 + {} - 1", kw), "Unexpected keyword in expression");
         }
     }
@@ -1400,22 +1738,19 @@ mod tests {
     #[test]
     fn test_while_empty() {
         do_ok_test(
-            "WHILE 2 + 3\nEND WHILE",
+            "WHILE 2 + 3\nWEND",
             &[Statement::While(
                 Expr::Add(Box::from(Expr::Integer(2)), Box::from(Expr::Integer(3))),
                 vec![],
             )],
         );
-        do_ok_test(
-            "WHILE 5\n\nREM foo\n\nEND WHILE\n",
-            &[Statement::While(Expr::Integer(5), vec![])],
-        );
+        do_ok_test("WHILE 5\n\nREM foo\n\nWEND\n", &[Statement::While(Expr::Integer(5), vec![])]);
     }
 
     #[test]
     fn test_while_loops() {
         do_ok_test(
-            "WHILE TRUE\nA\nB\nEND WHILE",
+            "WHILE TRUE\nA\nB\nWEND",
             &[Statement::While(
                 Expr::Boolean(true),
                 vec![
@@ -1433,9 +1768,9 @@ mod tests {
                 A
                 WHILE FALSE
                     B
-                END WHILE
+                WEND
                 C
-            END WHILE
+            WEND
         "#;
         do_ok_test(
             code,
@@ -1457,10 +1792,11 @@ mod tests {
     fn test_while_errors() {
         do_error_test("WHILE\n", "No expression in WHILE statement");
         do_error_test("WHILE TRUE", "Expecting newline after WHILE");
-        do_error_test("WHILE TRUE\n", "WHILE without END WHILE");
-        do_error_test("WHILE TRUE\nEND", "WHILE without END WHILE");
-        do_error_test("WHILE TRUE\nEND\n", "WHILE without END WHILE");
+        do_error_test("WHILE TRUE\n", "WHILE without WEND");
+        do_error_test("WHILE TRUE\nEND", "Unexpected token End in statement");
+        do_error_test("WHILE TRUE\nEND\n", "Unexpected token End in statement");
+        do_error_test("WHILE TRUE\nEND WHILE\n", "Unexpected token End in statement");
 
-        do_error_test("WHILE ,\nEND WHILE", "No expression in WHILE statement");
+        do_error_test("WHILE ,\nWEND", "No expression in WHILE statement");
     }
 }
